﻿--
-- constants
--
local SOUND_REQUEST=         "Sound\\interface\\iCreateCharacterA.wav";
local SOUND_FAILED=          "Sound\\interface\\igQuestFailed.wav";
local ICON_SHARD=            "Interface\\Icons\\INV_Misc_Gem_Amethyst_02";
local ICON_SPELL=            "Interface\\Icons\\Spell_Shadow_Twilight";
local SOULSHARD_ID=          6265;   -- soulshard item ID
local summonQueueMax=        15;     -- max. no. of players in queue
local summonMaxMsgLen=       30;     -- max. length of request message

--
-- default values
--
local realChatFrameOnEvent=  nil;    -- original versions of hooked functions
local realSetAction=         nil;
local realUseAction=         nil;
local isAddonLoaded=         false;  -- set after VARIABLES_LOADED
local summonQueue=           { };    -- table of queued players
local summonQueueCount=      0;
local isAFK=                 false;
local summoningCharacter=    nil;    -- name of player currently being summoned
local statusSummoning=       0;      -- 1: spellcast has started, 2: channeling has started, 3: casting has ended, 4: channeling is done
local shardsSummoning=       0;      -- shardcount before summoning: used to verfiy successful summon
local ghostNotify=           nil;    -- remember if target has already been notified that summoning ghosts isn't possible
local isMassSummoning=       false;  -- set while mass summoning a raid
local alreadySummoned=       { };    -- targets already summoned by mass summon
local registeredForUpdate=   { };    -- list of functions to call during UpdateInterfaces()

--
-- initialize saved states
--
rsm_SummonsList=             { };
rsm_Enabled=                 true;
rsm_Announce=                true;
rsm_Whisper=                 true;
rsm_Listening=               true;
rsm_SummonMinimum=           7;
rsm_SummonTL=                string.sub(GetLocale(), 1, 2);
rsm_Keywords=                { };
rsm_HideRequests=            false;
rsm_HideReplies=             true;


--
-- register events/command on addon load
--
function RSM_OnLoad()
  local _, unitClass= UnitClass("player");
  if (unitClass ~= "WARLOCK") then
    return;
  end;
  
  -- need to catch whispers, afk events and spellcast events
  this:RegisterEvent("VARIABLES_LOADED");
  this:RegisterEvent("CHAT_MSG_WHISPER");
  this:RegisterEvent("CHAT_MSG_SYSTEM");
  this:RegisterEvent("SPELLCAST_START");
  this:RegisterEvent("SPELLCAST_STOP");
  this:RegisterEvent("SPELLCAST_CHANNEL_START");
  this:RegisterEvent("SPELLCAST_CHANNEL_STOP");
  this:RegisterEvent("SPELLCAST_FAILED");
  this:RegisterEvent("SPELLCAST_INTERRUPTED");

  -- catch requests/replies before they reach the chatframe
  realChatFrameOnEvent=  ChatFrame_OnEvent;
  ChatFrame_OnEvent=     RSM_ChatFrameOnEvent;

  -- hook to replace normal "Ritual of Summoning" with our improved version
  realUseAction=    UseAction;
  UseAction=        RSM_UseAction;

  -- hooking this so we can enhance the tooltip for the ritual
  realSetAction=         GameTooltip.SetAction;
  GameTooltip.SetAction= RSM_SetAction;

  -- register mass summoning confirmation dialog
  StaticPopupDialogs["SUMMON_ALL"] = {
    text=      RSM_CONFIRMMASSSUMMON,
    button1=   TEXT(ACCEPT),
    button2=   TEXT(CANCEL),
    OnAccept=  RSM_SummonAll,
    showAlert= 2,
    timeout=   0
  };

  SlashCmdList["RAIDSUMMON"]= RSM_Command;
  SLASH_RAIDSUMMON1= "/raidsummon";
  SLASH_RAIDSUMMON2= "/rsm";
end;


--
-- simple support functions for text output
--
local function DisplayMessage(msg, isError)
  if (isError) then
    UIErrorsFrame:AddMessage(msg, 1.0, 0.0, 0.0, 1.0, 3);
  else
    UIErrorsFrame:AddMessage(msg, 1.0, 0.3, 1.0, 1.0, 3);
  end;
end;
local function PrintChatMessage(msg)
  DEFAULT_CHAT_FRAME:AddMessage(msg);
end;
local function PartyOrRaidMessage(msg)
  if (GetNumRaidMembers() > 0) then
    SendChatMessage(msg, "RAID");
  elseif (GetNumPartyMembers() > 0) then
    SendChatMessage(msg, "PARTY");
  end;
end;


--
-- get a localized string from the table
--
local function GetLocalized(s)
  local lstring;

   -- make sure table is loaded
   if (RSM_LANG ~= nil) then
     -- reset to english, if language invalid
     if (RSM_LANG[rsm_SummonTL] == nil) then
       rsm_SummonTL= "en";
     end;
     
     -- if table now valid return string - if found
     if (RSM_LANG[rsm_SummonTL] ~= nil) then
       lstring= RSM_LANG[rsm_SummonTL][s];
       if (lstring == nil) then
         lstring= "";
         PrintChatMessage("RaidSummon: Internal error: Invalid translation table entry.");
       end;
       return lstring;
     
     -- if table still invalid, something is really wrong
     else
       PrintChatMessage("RaidSummon: Internal error: Translation table corrupted!");
       return "";
       
     end;
   
   -- table is not loaded, something is really really wrong
   else
     PrintChatMessage("RaidSummon: Internal error: No translation table found!");
     return "";
     
   end;
end;


--
-- send localized tell to a player
-- replace %n with playername, %g{"a","b"} with genderspecific string "a" or "b"
--
local function SendTell(msgID, player)
  if (rsm_Whisper) then
    local msg= GetLocalized(msgID);
    msg= string.gsub(msg, "%%n", UnitName("player"));
    msg= string.gsub(msg, "%%g%{(%a+),(%a+)%}", "%"..(UnitSex("player") -1));
    SendChatMessage(msg, "WHISPER", this.language, player);
  end;
end;


--
-- check if item is a soulshard
--
local function IsSoulshard(bag, slot)
  local link= GetContainerItemLink(bag, slot);
  if (link) then
    local _,_, itemID= string.find(link, "item:(%d+):");
    if (itemID and tonumber(itemID) == SOULSHARD_ID) then
      return true;
    end;
  end;
  return false;
end;


--
-- count available shards (not perfect, should probably verify item link)
--
local function GetShardCount()
  local shardCount= 0;
  for bag= 4, 0, -1 do
    local size= GetContainerNumSlots(bag);
    if (size > 0) then
      for slot= 1, size, 1 do
        local texture= GetContainerItemInfo(bag, slot);
        if ((texture == ICON_SHARD) and IsSoulshard(bag, slot)) then
          shardCount= shardCount +1;
        end
      end
    end
  end
  return shardCount;
end;


--
-- return unit name of nickname
--
local function GetPID(nickname, idOnly)
  local param= "";
  local max=0;

  if (GetNumRaidMembers() > 0) then
    max= 40;
    param= "raid";
  elseif (GetNumPartyMembers() > 0) then
    max= 4;
    param= "party";
  end;

  if (param ~= "") then
    for i=1,max do
      if (UnitExists(param..i)) then
        if (nickname == UnitName(param..i)) then
          if (idOnly) then
            return i;
          else
            return param..i;
          end;
        end;
      end;
    end;
  end;

  return "";
end;


--
-- update interfaces
--
local function UpdateInterfaces()
  for _, func in registeredForUpdate do
    func();
  end;
end;


--
-- check if message is a port request
--
local function IsPortRequest(nickname, msg)
  if ((string.len(msg) < summonMaxMsgLen) and (GetPID(nickname) ~= "")) then
    local lowMsg= string.lower(msg);

    for i= 1, table.getn(rsm_Keywords) do
      if (string.find(lowMsg, string.lower(rsm_Keywords[i]))) then
        return true;
      end;
    end;
  end;
  
  return false;
end;


--
-- check if player already queued
--
local function CheckPlayerQueued(playerName)
  local i;
  for i= 1, summonQueueCount do
    if (summonQueue[i] == playerName) then 
      return true;
    end;
  end;
  return false;
end;


--
-- check if raidmembers location is same as players, requires raidindex, not unitid
--
local function SameLocation(id)
  local name, _, _, _, _, _, targetZone, isOnline= GetRaidRosterInfo(id);

  local myZone=     GetRealZoneText();
  local mySubZone=  GetSubZoneText();

  -- try to generalize zone names to find matching zones easier
  if (RSM_SubZones[targetZone]) then
    targetZone= RSM_SubZones[targetZone];
  end;
  if (RSM_SubZones[myZone]) then
    myZone= RSM_SubZones[myZone];
  end;
  if (RSM_SubZones[mySubZone]) then
    mySubZone= RSM_SubZones[mySubZone];
  end;

  -- don't summon people from same zone or instance or offline people
  if ((not isOnline) or
      (targetZone == myZone) or
      (targetZone == mySubZone))
  then
    return true;
  else
    return false;
  end;
end;


--
-- scan for members in different zones
--
local function FindDistantMember()
  local warlock=    nil;   -- contains name of a warlock that could be summoned, if any
  local member=     nil;   -- name of a raidmember that could be summoned, if any

  -- scan through all units
  for i=1, 40 do
    local unitID= "raid"..i;

    -- if unit exists, check if warlock and/or in same location
    if (UnitExists(unitID) and not UnitIsDeadOrGhost(unitID) and UnitIsConnected(unitID)) then
      local _, unitClass= UnitClass(unitID);

      if ((SameLocation(i) == false) and (alreadySummoned[UnitName(unitID)] ~= true)) then
        member= UnitName(unitID);

        -- if warlock, break here and summon
        if (unitClass == "WARLOCK") then
          warlock= UnitName(unitID);
          break;
        end;
      end;
    end;

  end;

  -- if warlock found, return name, otherwise return some members name
  if (warlock ~= nil) then
    return warlock;
  elseif (member ~= nil) then
    return member;
  end;
         
  -- if nothing found or not in raidgroup, return nil
  return nil;
end;


--
-- add player to summoning list
--
local function AddSummoning(nickname)
  -- check if warlock is AFK
  if (isAFK == true) then
    SendTell("AFK", nickname);
    return;
  end;

  -- ignore requests from already queued players
  if (CheckPlayerQueued(nickname)) then
    return;
  end;

  -- make sure enough shards remain once this player can be summoned
  if (GetShardCount() - summonQueueCount < rsm_SummonMinimum) then
    SendTell("SHARDS", nickname);
    return;
  end;

  -- make sure queue is not full
  if (summonQueueCount >= summonQueueMax) then
    SendTell("FULL", nickname);
    return;
  end;

  -- notify player if there are other people in the queue, player is in combat or target not in instance
  if (IsInInstance() and not SameLocation(GetPID(nickname, true))) then
    SendTell("NOT_INSTANCE", nickname);
  elseif ((UnitIsDeadOrGhost("player")) or (UnitAffectingCombat("player") == 1)) then
    SendTell("BUSY", nickname);
  elseif (summonQueueCount > 0) then
    SendTell("WAIT", nickname);
  else
    SendTell("INCOMING", nickname);
  end;

  -- notify warlock on first summon request
  if (summonQueueCount == 0) then
    PlaySoundFile(SOUND_REQUEST);
  end;

  -- otherwise add player to queue, visually notify warlock
  table.insert(summonQueue, nickname);
  summonQueueCount= summonQueueCount +1;
  UpdateInterfaces();

  DisplayMessage(RSM_REQUEST..nickname, false);
  PrintChatMessage(RSM_HEADER..RSM_REQUEST..nickname);
end;


--
-- update summoning stats
--
local function UpdateSummonStats(nickname)
  if (not rsm_SummonsList[nickname]) then
    rsm_SummonsList[nickname]= 1;
  else
    rsm_SummonsList[nickname]= rsm_SummonsList[nickname] +1;
  end;
end;


--
-- remove player playerName from queue
-- returns true if player was removed or false if not found
--
local function DequeuePlayer(playerName)
  local i;
  for i=1, summonQueueCount do
    if (summonQueue[i] == playerName) then
      table.remove(summonQueue, i);
      summonQueueCount= summonQueueCount -1;
      return true;
    end;
  end;
  return false;
end;


--
-- find player for mass summoning and add to list
--
local function AddMassSummoning()
  local nickname=    nil;
  local justStarted= not isMassSummoning;

  -- locate distant member and enqueue first
  nickname= FindDistantMember();
  if (nickname ~= nil) then

    -- set mass summoning flag, so other functions know what to do
    isMassSummoning= true;

    -- if target is already queued, remove. in any case add as first in queue
    DequeuePlayer(nickname);
    table.insert(summonQueue, 1, nickname);
    summonQueueCount= summonQueueCount +1;

    -- only print mass summoning notice once to chatlog
    if (justStarted == true) then
      PrintChatMessage(RSM_MASSSUMMON);
    end;

  -- no more members out of range
  else
    isMassSummoning= false;
    alreadySummoned= { };

    -- if just started and nothing to do notify user that everybody is in range
    if (justStarted == true) then
      PlaySoundFile(SOUND_FAILED);
      DisplayMessage(RSM_ALL_IN_RANGE, true);
      --PrintChatMessage(RSM_HEADER..RSM_ALL_IN_RANGE);

    -- otherwise notify user that we are done
    else
      DisplayMessage(RSM_NOTHING_TO_DO, false);
      --PrintChatMessage(RSM_HEADER..RSM_NOTHING_TO_DO);
    end;
  end;
end;


--
-- summon the player currently queued first
-- returns false if target was not changed
--
local function Summon()
  local unitname;
  local nickname;
  local shards= GetShardCount();

  -- return if nobody in queue
  if (summonQueueCount == 0) then
    DisplayMessage(RSM_NOTHING_TO_DO, true);
    --PrintChatMessage(RSM_HEADER..RSM_NOTHING_TO_DO);
    return false;
  end;

  -- select player
  nickname= summonQueue[1];
  unitname= GetPID(nickname);

  -- return if unit not found (i.e. summoning self outside of raid)
  if (unitname ~= "") then
    TargetUnit(unitname);
  else
    DisplayMessage(RSM_NOT_IN_RAID, true);
    --PrintChatMessage(RSM_HEADER..RSM_NOT_IN_RAID);
    DequeuePlayer(nickname);
    UpdateInterfaces();
    return false;
  end;

  -- check shards again (might have been used for stones or preferenced ports)
  if (shards == 0) then
    --PrintChatMessage(RSM_HEADER..RSM_NOSHARDSUMMONING..nickname);
    DisplayMessage(RSM_NOSHARDSUMMONING..nickname, true);
    PlaySoundFile(SOUND_FAILED);
      
  -- check if player is alive, then summon him/her and update stats
  elseif (UnitIsDeadOrGhost("target")) then
      
    -- if more players queued, skip dead guy for now and re-enqueue at end of list
    if (summonQueueCount > 1) then
      table.remove(summonQueue, 1);
      table.insert(summonQueue, nickname);
      PrintChatMessage(RSM_HEADER..RSM_GHOSTSKIPPED..nickname);
      DisplayMessage(RSM_GHOSTSKIPPED..nickname, false);
        
    else
      -- notify player one time that he cannot be summoned
      if (ghostNotify ~= nickname) then
        SendTell("YOU_ARE_DEAD", nickname);
        ghostNotify= nickname;
      end;
      PrintChatMessage(RSM_HEADER..RSM_GHOSTSUMMONING..nickname);
      DisplayMessage(RSM_GHOSTSUMMONING..nickname, true);
      PlaySoundFile(SOUND_FAILED);        

    end;       


  else
    -- remember that we're currently summoning and how much shards we had
    -- tracking stats and notifying target is done in event handler for spellcast start
    summoningCharacter= nickname;
    shardsSummoning= shards;
    CastSpellByName(RSM_SPELL_NAME);

  end;

  -- refresh interfaces
  UpdateInterfaces();
  return true;
end;


--
-- initialize RaidSummon (called once variables are loaded)
--
local function Initialize()
  local _= GetLocalized("LANGUAGE"); -- will correct invalid language setting
  if (table.getn(rsm_Keywords) == 0) then
    for i, keyword in RSM_DEFAULT_KEYWORDS do
      rsm_Keywords[i]= keyword;
    end;
  end;
  UpdateInterfaces();
  PrintChatMessage(RSM_LOADED);
  isAddonLoaded= true;
end;


--
-- update info about players AFK status
--
local function UpdatePlayerAFK(msg)
  local afkstring= string.format(MARKED_AFK_MESSAGE, "");
  if (string.find(msg, afkstring)) then
    isAFK= true;
    UpdateInterfaces();
  elseif (msg == CLEARED_AFK) then
    isAFK= false;
    UpdateInterfaces();
  end;
end;


--
-- handle game-events
--
function RSM_OnEvent(event)
  -- check initialization state, init if neccessary
  if (event == "VARIABLES_LOADED") then
    Initialize();
  end;
  if (not isAddonLoaded) then
    return;
  end;

  -- handle port request
  if (event == "CHAT_MSG_WHISPER") then
    -- arg1: message
    -- arg2: nickname
    if (rsm_Enabled and rsm_Listening and IsPortRequest(arg2, arg1)) then
      AddSummoning(arg2);
    end;

  -- detect AFK
  elseif ((event == "CHAT_MSG_SYSTEM") and (arg1 ~= nil)) then
    -- arg1: message
    UpdatePlayerAFK(arg1)
    
  end;

  -- monitor summoning progress
  if (summoningCharacter ~= nil) then

    -- verify that summoning actually started (might not have worked if sitting, oom, oos...)
    if ((event == "SPELLCAST_START") and (statusSummoning == 0)) then
      statusSummoning= 1;
      PrintChatMessage(RSM_HEADER..RSM_SUMMONING..summoningCharacter);
      DisplayMessage(RSM_SUMMONING..summoningCharacter, false);
      if (rsm_Announce) then
        PartyOrRaidMessage(string.format(GetLocalized("ANNOUNCE_PORT"), summoningCharacter));
      end;

    -- channeling has started, notify players to assist and notify target
    elseif ((event == "SPELLCAST_CHANNEL_START") and (statusSummoning == 1)) then
      statusSummoning= 2;
      SendTell("NOW", summoningCharacter);
      if (rsm_Announce) then
        PartyOrRaidMessage(GetLocalized("PORT_SUPPORT"));
      end;

    -- casting has ended
    elseif ((event == "SPELLCAST_STOP") and (statusSummoning == 2)) then
      statusSummoning= 3;

    -- channeling has ended, check if shard was used
    elseif ((event == "SPELLCAST_CHANNEL_STOP") and (statusSummoning == 3)) then
      if (GetShardCount() < shardsSummoning) then
        UpdateSummonStats(summoningCharacter);

        -- remove target from request queue
        table.remove(summonQueue, 1);
        summonQueueCount= summonQueueCount -1;

        -- if currently mass summoning, aquire next target
        if (isMassSummoning == true) then
          alreadySummoned[summoningCharacter]= true;
          AddMassSummoning();
        end;

        UpdateInterfaces();

        -- alert player that more summoning requests are queued
        if (summonQueueCount > 0) then
          DisplayMessage(RSM_MORE..summonQueue[1], false);
        end;

      else
        PlaySoundFile(SOUND_FAILED);
        DisplayMessage(RSM_FAILED, true);
        PrintChatMessage(RSM_HEADER..RSM_FAILED);
        if (rsm_Announce) then
          PartyOrRaidMessage(GetLocalized("PORT_FAILED"));
        end;

      end;

      summoningCharacter= nil;
      statusSummoning= 0;

    -- failure
    elseif ((event == "SPELLCAST_FAILED") or (event == "SPELLCAST_INTERRUPTED")) then
      summoningCharacter= nil;
      statusSummoning= 0;

      PlaySoundFile(SOUND_FAILED);
      DisplayMessage(RSM_FAILED, true);
      PrintChatMessage(RSM_HEADER..RSM_FAILED);

    end;

  end;
end;


--
-- hook chatframe-events: prevent port request from spamming the chat
--
function RSM_ChatFrameOnEvent(event, ...)
  -- arg1: message
  -- arg2: name

  if (rsm_Enabled and RSM_PREFIX) then
    if (rsm_HideReplies and event == "CHAT_MSG_WHISPER_INFORM" and (string.find(arg1, RSM_PREFIX, 1, true) == 1)) then
      return;
    elseif (rsm_HideRequests and event == "CHAT_MSG_WHISPER" and IsPortRequest(arg2, arg1)) then
      return;
    end;
  end;

  return realChatFrameOnEvent(event, unpack(arg));
end;


-- ******************************************************************
--                       UseAction interfaces
-- ******************************************************************

--
-- add menu command/toggle/title/spacer
--
local function AddCommand(text, enabled, func)
  local info= { };
  info.text=     text;
  info.disabled= not enabled;
  info.func=     func;
  UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
end;
local function AddToggle(text, enabled, getfunc, togglefunc)
  local info= { };
  info.text=             text;
  info.disabled=         not enabled;
  info.func=             togglefunc;
  info.checked=          getfunc();
  info.keepShownOnClick= 1;
  UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
end;
local function AddTitle(title)
  local info= {};
  info.text=         title;
  info.notClickable= 1;
  info.isTitle=      1;
  UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
end;
local function AddSpacer()
  local info= {};
  info.disabled= 1;
  UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
end;


--
-- it's rather unlikely that there is really no function in the
-- default UI which does this, but I didn't find it.
--
local function SelectLang()
  local level= 2;

  if (this.value) then
    RSM_SetNotificationLanguage(this.value, false);

    -- re-enable and deselect other language buttons
    local listFrame= getglobal("DropDownList"..level);
    for i= 1, listFrame.numButtons do
      local button=          getglobal("DropDownList"..level.."Button"..i);
      local invisibleButton= getglobal("DropDownList"..level.."Button"..i.."InvisibleButton");
      local checkImage=      getglobal("DropDownList"..level.."Button"..i.."Check");
      if (button.value and (type(button.value) == "string") and RSM_LANG[button.value]) then
        button.checked= nil;
        button.notClickable= nil;
        button:Enable();
        invisibleButton:Hide();
        checkImage:Hide();
      end;
    end;

    -- disable this button
    this:SetDisabledTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b);
    local invisibleButton= getglobal(this:GetName().."InvisibleButton");
    this:Disable();
    invisibleButton:Show();
  end;    
end;
local function SelectShards()
  local level= 2;

  if (this.value) then
    RSM_SetShardMinimum(this.value, false);

    -- re-enable and deselect other shardcount buttons
    local listFrame= getglobal("DropDownList"..level);
    for i= 1, listFrame.numButtons do
      local button=          getglobal("DropDownList"..level.."Button"..i);
      local invisibleButton= getglobal("DropDownList"..level.."Button"..i.."InvisibleButton");
      local checkImage=      getglobal("DropDownList"..level.."Button"..i.."Check");
      if (button.value and tonumber(button.value)) then
        button.checked= nil;
        button.notClickable= nil;
        button:Enable();
        invisibleButton:Hide();
        checkImage:Hide();
      end;
    end;

    -- disable this button
    this:SetDisabledTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b);
    local invisibleButton= getglobal(this:GetName().."InvisibleButton");
    this:Disable();
    invisibleButton:Show();
  end;    
end;


--
-- create basic menu entries
--
function RSM_BuildBaseMenu()
  -- create notification submenu
  if (UIDROPDOWNMENU_MENU_LEVEL == 2 and this.value == "notifications") then
    AddTitle(RSM_MENU_SELECT);
    AddToggle(RSM_MENU_ANNOUNCING, true, RSM_GetAnnouncing, RSM_ToggleAnnounce);
    AddToggle(RSM_MENU_WHISPERING, true, RSM_GetWhispering, RSM_ToggleWhisper);
    AddSpacer();

    AddTitle(RSM_MENU_DISPLAY);
    AddToggle(RSM_MENU_HIDE_REQUESTS, true, RSM_GetHideRequests, RSM_ToggleHideRequests);
    AddToggle(RSM_MENU_HIDE_TELLS,    true, RSM_GetHideReplies,  RSM_ToggleHideReplies);
    AddSpacer();

    AddTitle(RSM_MENU_LOCALE);
    local curLang= RSM_GetNotificationLanguage();
    for tag, data in RSM_LANG do
      local name= data["LANGUAGE"];
      local info= { };
      info.text=             name;
      info.checked=          (tag == curLang);
      info.notClickable=     (tag == curLang);
      info.keepShownOnClick= 1;
      info.value=            tag;
      info.func=             SelectLang;
      UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
    end;

  -- create minumum-shard submenu
  elseif (UIDROPDOWNMENU_MENU_LEVEL == 2 and this.value == "shards") then
    AddTitle(RSM_MENU_SHARDS);
    local curMinimum= RSM_GetShardMinimum();
    for i= 1, 28, 3 do
      local info= { };
      info.text=             i;
      info.checked=          (i == curMinimum);
      info.notClickable=     (i == curMinimum);
      info.keepShownOnClick= 1;
      info.value=            i;
      info.func=             SelectShards;
      UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
    end;

  elseif (UIDROPDOWNMENU_MENU_LEVEL == 1) then
    -- check current environment
    local inRaid=         (GetNumRaidMembers() > 0);
    local inGroup=        (GetNumPartyMembers() > 0);
    local targetInRaid=   (UnitInParty("target") or UnitInRaid("target"));
    local somebodyQueued= (RSM_GetSummonQueueCount() ~= 0);
    local massSummoning=  RSM_GetSummonAll();
    local skipText;
    if (massSummoning) then
      skipText= RSM_MENU_STOP;
    else
      skipText= RSM_MENU_CLEAR;
    end;

    -- prevent stupid submenu arrows from actually hiding the submenu they are supposed to open... congrats Blizzard!
    for i= 1, UIDROPDOWNMENU_MAXBUTTONS do
      getglobal("DropDownList"..UIDROPDOWNMENU_MENU_LEVEL.."Button"..i.."ExpandArrow"):Disable();
    end;

    -- create menu
    AddTitle(BINDING_HEADER_RAIDSUMMON);

    AddCommand(RSM_MENU_SUMMON,        (inGroup or inRaid) and somebodyQueued,              RSM_SummonQueued);
    AddCommand(RSM_MENU_SUMMON_TARGET, (inGroup or inRaid) and targetInRaid,                RSM_SummonTarget);
    AddCommand(RSM_MENU_SUMMON_ALL,    inRaid and not massSummoning and not IsInInstance(), RSM_SummonAll);
    AddSpacer();

    AddCommand(RSM_MENU_SKIP, somebodyQueued, RSM_SkipQueue);
    AddCommand(skipText, somebodyQueued, RSM_ResetQueue);
    AddCommand(RSM_MENU_ANNOUNCE, inGroup or inRaid, RSM_Announce);
    AddSpacer();

    AddToggle(RSM_MENU_PROCESS, true, RSM_GetListening, RSM_ToggleListen);
    UIDropDownMenu_AddButton({ ["text"]= RSM_MENU_SHARD_MINIMUM, ["hasArrow"]= 1, ["value"]="shards" });
    UIDropDownMenu_AddButton({ ["text"]= RSM_MENU_NOTIFICATIONS, ["hasArrow"]= 1, ["value"]="notifications" });

  end;
end;


--
-- close dropdownmenu
--
local function CloseActionMenu()
  DropDownList1:Hide();
end;


--
-- extend basemenu with action-specific entries
--
function RSM_BuildActionMenu()
  RSM_BuildBaseMenu();
  if (UIDROPDOWNMENU_MENU_LEVEL == 1) then
    AddSpacer();
    AddCommand(RSM_MENU_CLOSE, true, CloseActionMenu);
  end;
end;


--
-- hook action tooltip to enhance "Ritual of Summoning" tooltip
--
function RSM_SetAction(self, id, ...)
  local rettab= { realSetAction(self, id, unpack(arg)) };

  if (rsm_Enabled) then
    local isMacro=   (GetActionText(id) ~= nil);
    local texture=   GetActionTexture(id);
    local spellName= GameTooltipTextLeft1:GetText();
    local queue;

    -- check if selected action is a ritual of summoning
    if (not isMacro and spellName == RSM_SPELL_NAME and texture == ICON_SPELL) then
      if (tonumber(GetCVar("UberTooltips")) == 1) then
        GameTooltipTextLeft5:SetText(RSM_SPELL_HELP.."\n\n");
      end;
    
      if (summonQueueCount == 0) then
        queue= NORMAL_FONT_COLOR_CODE..RSM_SPELL_QUEUE_EMPTY..FONT_COLOR_CODE_CLOSE.."\n";
      else
        queue= NORMAL_FONT_COLOR_CODE..RSM_SPELL_QUEUE..FONT_COLOR_CODE_CLOSE;
        for i=1, summonQueueCount -1 do
          queue= queue..summonQueue[i]..", ";
        end;
        queue= queue..summonQueue[summonQueueCount]..".";
      end;
      GameTooltip:AddLine(queue, 1.0, 1.0, 1.0, 1);
      GameTooltip:Show();
    end;
  end;

  return unpack(rettab);
end;


--
-- hook UseAction function to detect when the player is about to cast "Ritual of Summoning"
--
function RSM_UseAction(id, ex, onSelf, ...)
  -- return if addon disabled or drag and drop action taking place (cannot check for macro drag&drop)
  if (not rsm_Enabled or CursorHasItem() or CursorHasSpell()) then
    return realUseAction(id, ex, onSelf, unpack(arg));
  end;

  -- retrieve tooltip data
  RSMTooltip:SetOwner(UIParent, "ANCHOR_NONE");
  RSMTooltip:SetAction(id);
  local spellName= RSMTooltipTextLeft1:GetText();
  local isMacro=   (GetActionText(id) ~= nil);
  local texture=   GetActionTexture(id);

  -- if it's the ritual, cast it, otherwise use default action
  if (not isMacro and spellName == RSM_SPELL_NAME and texture == ICON_SPELL) then

    -- if clicked with right mouse, show dropdown menu
    if (arg1 and arg1 == "RightButton") then
      local x, y= GetCursorPosition();
      GameTooltip:Hide();
      ToggleDropDownMenu(1, nil, RSM_DropDownMenu, "RSM_DropDownMenu", x, y);
      return;

    else
      CloseActionMenu();
      RSM_PerformSummoning();
 
    end;

  else
    return realUseAction(id, ex, onSelf, unpack(arg));

  end;
end;


-- ******************************************************************
--                       public interfaces
-- ******************************************************************


--
-- register function to be called on interface update
--
function RSM_RegisterForUpdate(func)
  if (type(func) == "function") then
    table.insert(registeredForUpdate, func);
  end;
end;


--
-- semi-intelligent summoning mode autoselection
--
function RSM_PerformSummoning()
  local isGrouped= (GetNumRaidMembers() > 0) or (GetNumPartyMembers() > 0);

  -- skip player if control key down
  if (IsControlKeyDown()) then
    RSM_SkipQueue();

  -- if not in a group/raid, tell player
  elseif (not isGrouped) then
    DisplayMessage(RSM_GROUP_REQUIRED, true);

  -- player has friendly target, summon it
  elseif (UnitExists("target") and UnitIsFriend("player", "target")) then
    RSM_SummonTarget(); -- will check if target in group

  -- other players are queued, summon them
  elseif (summonQueueCount > 0) then
    RSM_SummonQueued();

  -- nobody queued and in raid: summon everybody
  elseif ((GetNumRaidMembers() > 0) and StaticPopupDialogs["SUMMON_ALL"]) then
    if (IsInInstance()) then
      DisplayMessage(RSM_NOT_IN_INSTANCE, true);
    else
      StaticPopup_Show("SUMMON_ALL");
    end;

  -- no friendly target, nobody queued, not in raid but grouped: must be in party
  else
    DisplayMessage(RSM_NOT_IN_RAID, true);

  end;
end;


--
-- Summon first player in queue (also hiding targetchange from player)
--
function RSM_SummonQueued()
  local nullTarget;

  -- return if already summoning
  if (summoningCharacter ~= nil) then
    return;
  end;
  
  -- remember if something was targetted before
  if (UnitName("target") == nil) then
    nullTarget= true;
  else
    nullTarget= false; 
  end;

  -- summon player
  local changedTarget= Summon();

  -- restore previous target, if any, and only if it was changed
  if (changedTarget == true) then
    local lala= UnitName("target");
    if (nullTarget == true) then
      ClearTarget();
    else
      TargetLastTarget();
    end;
  end;
end;


--
-- Summon everybody not currently in this zone
--
function RSM_SummonAll()
  if (GetNumRaidMembers() == 0) then
    DisplayMessage(RSM_NO_RAID, true);
    --PrintChatMessage(RSM_HEADER..RSM_NO_RAID);
    return;
  elseif (IsInInstance()) then
    DisplayMessage(RSM_NOT_IN_INSTANCE, true);
    --PrintChatMessage(RSM_HEADER..RSM_NOT_IN_INSTANCE);
    return;
  end;

  AddMassSummoning();

  if (isMassSummoning) then
    RSM_SummonQueued();
  end;
end;


--
-- Summon target (by enqueing player first)
--
function RSM_SummonTarget()
  -- return if already summoning
  if (summoningCharacter ~= nil) then
    return;
  end;

  -- check if target is in raid or group. 
  if (not (UnitInParty("target") or UnitInRaid("target"))) then
    DisplayMessage(RSM_NOT_IN_RAID, true);
    --PrintChatMessage(RSM_HEADER..RSM_NOT_IN_RAID);
    return;
  end;

  -- if target is already queued, remove. in any case add as first in queue
  nickname= UnitName("target");
  DequeuePlayer(nickname);
  table.insert(summonQueue, 1, nickname);
  summonQueueCount= summonQueueCount +1;

  -- target is queued, now summon it just like a normal queued raidmember
  Summon();
end;


--
-- reset summoning queue
--
function RSM_ResetQueue()
  summonQueue= { };
  summonQueueCount= 0;
  isMassSummoning= false;
  alreadySummoned= { };

  UpdateInterfaces();
  PrintChatMessage(RSM_RESET_QUEUE);
end;


--
-- remove first queued player from list and notify player
--
function RSM_SkipQueue()
  local nickname;
  if (summonQueueCount > 0) then
    nickname= summonQueue[1];

    table.remove(summonQueue, 1);
    summonQueueCount= summonQueueCount -1;

    if (isMassSummoning) then
      alreadySummoned[nickname]= true;
      AddMassSummoning();
    end;

    UpdateInterfaces();
    PrintChatMessage(RSM_DEQUEUE_PLAYER..nickname);
    if (not isMassSummoning) then -- the message is "your request..." and there was no request in mass summoning mode
      SendTell("SKIP", nickname);
    end;
  end;
end;


--
-- enable/disable raidsummon functionality
--
function RSM_ToggleEnable(verbose)
  rsm_Enabled= not rsm_Enabled;
  if (verbose) then
    if (rsm_Enabled) then
      PrintChatMessage(RSM_ADDON_ENABLED);
    else
      PrintChatMessage(RSM_ADDON_DISABLED);
    end;
  end;
  UpdateInterfaces();
end;


--
-- enable/disable request processing
--
function RSM_ToggleListen(verbose)
  rsm_Listening= not rsm_Listening;
  if (verbose) then
    if (rsm_Listening) then
      PrintChatMessage(RSM_LISTEN_ENABLED);
    else
      PrintChatMessage(RSM_LISTEN_DISABLED);
    end;
  end;
  UpdateInterfaces();
end;


--
-- enable/disable summoning announcements
--
function RSM_ToggleAnnounce(verbose)
  rsm_Announce= not rsm_Announce;
  if (verbose) then
    if (rsm_Announce) then
      PrintChatMessage(RSM_ANNOUNCE_ENABLED);
    else
      PrintChatMessage(RSM_ANNOUNCE_DISABLED);
    end;
  end;
end;


--
-- enable/disable summoning notification by whisper
--
function RSM_ToggleWhisper(verbose)
  rsm_Whisper= not rsm_Whisper;
  if (verbose) then
    if (rsm_Whisper) then
      PrintChatMessage(RSM_WHISPER_ENABLED);
    else
      PrintChatMessage(RSM_WHISPER_DISABLED);
    end;
  end;
end;


--
-- show/hide requests
--
function RSM_ToggleHideRequests(verbose)
  rsm_HideRequests= not rsm_HideRequests;
  if (verbose) then
    if (rsm_HideRequests) then
      PrintChatMessage(RSM_REQUESTS_HIDDEN);
    else
      PrintChatMessage(RSM_REQUESTS_SHOWN);
    end;
  end;
end;


--
-- show/hide RaidSummon's replies to request
--
function RSM_ToggleHideReplies(verbose)
  rsm_HideReplies= not rsm_HideReplies;
  if (verbose) then
    if (rsm_HideReplies) then
      PrintChatMessage(RSM_REPLIES_HIDDEN);
    else
      PrintChatMessage(RSM_REPLIES_SHOWN);
    end;
  end;
end;


--
-- select announcement/whisper language
--
function RSM_SetNotificationLanguage(newLang, verbose)
  rsm_SummonTL= newLang;
  local langName= GetLocalized("LANGUAGE");  -- this will also reset language if invalid, so we don't need to check this here

  if (verbose) then
    PrintChatMessage(RSM_WHISPERLANG..langName.." ("..rsm_SummonTL..")");
  end;
end;


--
-- set minimum number of shards required to process requests
--
function RSM_SetShardMinimum(num, verbose)
  if (num) then
    num= tonumber(num);
  end;
  if (num and num >= 1) then
    rsm_SummonMinimum= num;
  else
    rsm_SummonMinimum= 1;
  end;

  if (verbose) then
    PrintChatMessage(RSM_MINIMUM..rsm_SummonMinimum);
  end;
end;


--
-- external interfaceing support: queue information
--
function RSM_GetSummonQueueCount()
  return summonQueueCount;
end;
function RSM_GetSummonQueue(index)
  return summonQueue[index];
end;


--
-- query status
--
function RSM_GetHideRequests()
  return rsm_HideRequests;
end;

function RSM_GetHideReplies()
  return rsm_HideReplies;
end;

function RSM_GetSummonAll()
  return isMassSummoning;
end;

function RSM_GetAFK()
  return isAFK;
end;

function RSM_GetEnabled()
  return rsm_Enabled;
end;

function RSM_GetListening()
  return rsm_Listening;
end;

function RSM_GetAnnouncing()
  return rsm_Announce;
end;

function RSM_GetWhispering()
  return rsm_Whisper;
end;

function RSM_GetNotificationLanguage()
  return rsm_SummonTL;
end;

function RSM_GetShardMinimum()
  return rsm_SummonMinimum;
end;


-- ******************************************************************
--                      command line support
-- ******************************************************************


--
-- list players in queue
--
function RSM_ListQueuePlayers()
  local i;
  PrintChatMessage(RSM_QUEUED_PLAYERS..summonQueueCount);
  for i= 1, summonQueueCount do
    PrintChatMessage(i..": "..summonQueue[i]);
  end;
  PrintChatMessage(RSM_END_OF_LIST);
end;
 
 
--
-- announce Summoning to Raid
--
function RSM_Announce()
  if ((GetNumRaidMembers() == 0) and (GetNumPartyMembers() == 0)) then
    PrintChatMessage(RSM_HEADER..RSM_NO_RAID);
    return
  end;

  PartyOrRaidMessage(GetLocalized("ANNOUNCE1"));
  PartyOrRaidMessage(GetLocalized("ANNOUNCE2"));
  PartyOrRaidMessage(GetLocalized("ANNOUNCE3"));
  PartyOrRaidMessage(GetLocalized("ANNOUNCE4"));
end;


--
-- print current settings
--
function RSM_PrintStatus()
  if (rsm_Enabled) then
    PrintChatMessage(RSM_ADDON_ENABLED);
  else
    PrintChatMessage(RSM_ADDON_DISABLED);
  end;
  if (rsm_Listening) then
    PrintChatMessage(RSM_LISTEN_ENABLED);
  else
    PrintChatMessage(RSM_LISTEN_DISABLED);
  end;
  if (rsm_Announce) then
    PrintChatMessage(RSM_ANNOUNCE_ENABLED);
  else
    PrintChatMessage(RSM_ANNOUNCE_DISABLED);
  end;
  if (rsm_Whisper) then
    PrintChatMessage(RSM_WHISPER_ENABLED);
  else
    PrintChatMessage(RSM_WHISPER_DISABLED);
  end;
  if (rsm_HideRequests) then
    PrintChatMessage(RSM_REQUESTS_HIDDEN);
  else
    PrintChatMessage(RSM_REQUESTS_SHOWN);
  end;
  if (rsm_HideReplies) then
    PrintChatMessage(RSM_REPLIES_HIDDEN);
  else
    PrintChatMessage(RSM_REPLIES_SHOWN);
  end;
  PrintChatMessage(RSM_WHISPERLANG..GetLocalized("LANGUAGE").." ("..rsm_SummonTL..")");
  PrintChatMessage(RSM_MINIMUM..rsm_SummonMinimum);
  PrintChatMessage(RSM_KEYWORDS..table.concat(rsm_Keywords, ", "));
end;


--
-- display/post summoning stats (FIXME: *still* pretty cheap implementation)
--
function RSM_PostRaidStats(maxCount)
  if ((GetNumRaidMembers() == 0) and (GetNumPartyMembers() == 0)) then
    PrintChatMessage(RSM_HEADER..RSM_NO_RAID);
    return
  end;

  PartyOrRaidMessage(GetLocalized("TOTALS_RAID"));

  local maxSummons= 0;
  for player, summons in rsm_SummonsList do
    if (summons >= maxSummons) then
      maxSummons= summons;
    end;
  end;

  local count= 0;
  local pleaseBreak= false;
  for i= maxSummons, 0, -1 do
    for player, summons in rsm_SummonsList do
      if (summons == i and GetPID(player) ~= "") then
        PartyOrRaidMessage(player..": "..summons);
        count= count +1;
        if (count >= maxCount) then
          pleaseBreak= true;
          break;
        end;
      end;
    end;
    if (pleaseBreak == true) then
      break;
    end;
  end;

  PartyOrRaidMessage(GetLocalized("END_OF_LIST"));
end;

function RSM_RaidStats(maxCount)
  if ((GetNumRaidMembers() == 0) and (GetNumPartyMembers() == 0)) then
    PrintChatMessage(RSM_HEADER..RSM_NO_RAID);
    return
  end;

  PrintChatMessage(RSM_TOTALS_RAID);

  local maxSummons= 0;
  for player, summons in rsm_SummonsList do
    if (summons >= maxSummons) then
      maxSummons= summons;
    end;
  end;

  local count= 0;
  local pleaseBreak= false;
  for i= maxSummons, 0, -1 do
    for player, summons in rsm_SummonsList do
      if (summons == i and GetPID(player) ~= "") then
        PrintChatMessage(player..": "..NORMAL_FONT_COLOR_CODE..summons..FONT_COLOR_CODE_CLOSE);
        count= count +1;
        if (count >= maxCount) then
          pleaseBreak= true;
          break;
        end;
      end;
    end;
    if (pleaseBreak == true) then
      break;
    end;
  end;

  PrintChatMessage(RSM_END_OF_LIST);
end;

function RSM_Stats(maxCount)
  local total= 0;
  local maxSummons= 0;
  for _, summons in rsm_SummonsList do
    total= total +summons;
    if (summons >= maxSummons) then
      maxSummons= summons;
    end;
  end;

  PrintChatMessage(RSM_TOTALS..NORMAL_FONT_COLOR_CODE..total..FONT_COLOR_CODE_CLOSE);

  local count= 0;
  local pleaseBreak= false;

  for i= maxSummons, 0, -1 do
    for player, summons in rsm_SummonsList do
      if (summons == i) then
        PrintChatMessage(player..": "..NORMAL_FONT_COLOR_CODE..summons..FONT_COLOR_CODE_CLOSE);
        count= count +1;
        if (count >= maxCount) then
          pleaseBreak= true;
          break;
        end;
      end;
    end;
    if (pleaseBreak == true) then
      break;
    end;
  end;

  PrintChatMessage(RSM_END_OF_LIST);  
end;


--
-- console callback
--
function RSM_Command(msg)
  -- only continue if already loaded
  if (isAddonLoaded == false) then
    return;
  end;
  
  -- process console commands
  if (msg == "") then
    PrintChatMessage(RSM_HELP_MSG2);
    PrintChatMessage(RSM_HELP_MSG17);
    PrintChatMessage(RSM_HELP_MSG15);
    PrintChatMessage(RSM_HELP_MSG8);
    PrintChatMessage(RSM_HELP_MSG16);
    PrintChatMessage(RSM_HELP_MSG7);
    PrintChatMessage(RSM_HELP_MSG10);
    PrintChatMessage(RSM_HELP_MSG13);
    PrintChatMessage(RSM_HELP_MSG12);
    PrintChatMessage(RSM_HELP_MSG18);
    PrintChatMessage(RSM_HELP_MSG14);
    PrintChatMessage(RSM_HELP_MSG9);
    PrintChatMessage(RSM_HELP_MSG11);
    PrintChatMessage(RSM_HELP_MSG6);
    PrintChatMessage(RSM_HELP_MSG3);
    PrintChatMessage(RSM_HELP_MSG4);
    PrintChatMessage(RSM_HELP_MSG5);
    PrintChatMessage(RSM_HELP_MSG19);
    PrintChatMessage(RSM_HELP_MSG1);
    PrintChatMessage(RSM_HELP_MSG20);

--[[
  elseif (msg == "debug") then
    summonQueueCount= 6;
    table.insert(summonQueue, "Lidee");
    table.insert(summonQueue, "Linea");
    table.insert(summonQueue, "Lisbeth");
    table.insert(summonQueue, "Lina");
    table.insert(summonQueue, "Lisanne");
    table.insert(summonQueue, "Lileile");
    UpdateInterfaces();
]]--
    
  elseif (msg == "status") then
    RSM_PrintStatus();

  elseif (msg == "resetstats") then
    rsm_SummonsTotal= 0;
    rsm_SummonsList= { };
    PrintChatMessage(RSM_STATS_RESET);

  elseif (string.find(msg, "postraidstats")) then
    local count= string.sub(msg, 15);
    if (count) then
      count= tonumber(count);
    end;
    if (count == nil or count == 0) then
      count= 10;
    end;
    RSM_PostRaidStats(count);

  elseif (string.find(msg, "raidstats")) then
    local count= string.sub(msg, 11);
    if (count) then
      count= tonumber(count);
    end;
    if (count == nil or count == 0) then
      count= 10;
    end;
    RSM_RaidStats(count);

  elseif (string.find(msg, "stats")) then
    local count= string.sub(msg, 7);
    if (count) then
      count= tonumber(count);
    end;
    if (count == nil or count == 0) then
      count= 10;
    end;
    RSM_Stats(count);

  elseif (msg == "clear") then
    RSM_ResetQueue();
  elseif (msg == "list") then
    RSM_ListQueuePlayers();
  elseif (msg == "announce") then
    RSM_Announce();
  elseif (msg == "toggleenable" or msg == "te") then
    RSM_ToggleEnable(true);
  elseif (msg == "togglelisten" or msg == "tl") then
    RSM_ToggleListen(true);
  elseif (msg == "toggleannounce" or msg == "ta") then
    RSM_ToggleAnnounce(true);
  elseif (msg == "togglewhisper" or msg == "tw") then
    RSM_ToggleWhisper(true);
  elseif (msg == "togglehiderequests" or msg == "thq") then
    RSM_ToggleHideRequests(true);
  elseif (msg == "togglehidereplies" or msg == "thp") then
    RSM_ToggleHideReplies(true);

  elseif (msg == "skip") then
    RSM_SkipQueue();

  -- FIXME: to be removed in next version
  elseif (msg == "process") then
    PrintChatMessage("RaidSummon no longer requires this macro. Please use the normal \"Ritual of Summoning\" spell instead.");
    PrintChatMessage("RaidSummon verwendet nun anstatt dieses Makros den Zauber \"Ritual der Besch\195\182rung\" direkt.");

  elseif (msg == "resetkeywords") then
    rsm_Keywords= {  };
    PrintChatMessage(RSM_KEYWORDRESET);

  elseif (string.find(msg, "addkeyword ") == 1) then
    local keyword= string.sub(msg, 12);
    if (keyword ~= "") then
      table.insert(rsm_Keywords, keyword);
    end;
    PrintChatMessage(RSM_KEYWORDS..table.concat(rsm_Keywords, ", "));

  elseif (string.find(msg, "minimum ") == 1) then
    local minimum= string.sub(msg, 9);
    RSM_SetShardMinimum(minimum, true);

  elseif (string.find(msg, "lang ") == 1) then
    local langAbr= string.sub(msg, 6);
    if (langAbr ~= "") then
      RSM_SetNotificationLanguage(langAbr, true);
    end;
    
  else
    PrintChatMessage(RSM_UNKNOWN_PARAMETER);

  end;

end;
